"""
class to word classification
"""
import re
import networkx as nx
from DST.utils.DSUtil import levenshtein_distance
import logging

logger = logging.getLogger(__name__)

def __numberInString(term):
    for i in term:
        if i.isdigit():
            return True
    return False


# check if the number in two string are the same
def __isNumberSame(term1, term2):
    # if one term has number inside, the other one should have the same number
    if __numberInString(term1) or __numberInString(term2):
        # if the number lists are different, they are not abbreviation relationships
        if re.findall(r"\d+", term1) != re.findall(r"\d+", term2):
            return False
    return True


# check if the letter order of two terms are the same
def __checkLetterOrder(shortTerm, longTerm):
    position = 0  # record the position of letter
    matchCount = 0  # how many letters in the short term are ordered as those in long term
    # do not take "-" and "_" into consideration
    for letter in shortTerm:
        addPosition = longTerm[position:].find(letter)
        if addPosition == -1:
            break
        else:
            matchCount = matchCount + 1
        position = position + addPosition + 1
    if matchCount == len(shortTerm):
        # if the abbreviation only refer to the the first word in long term, it is not the abbreviation of the whole term such as advace --> advace_mike
        # the last position of the blank i.e., the letters in an abbreviation should lay in all components in the words
        if " " in longTerm and position < longTerm.rfind(" ") + 1 and (shortTerm[-1] != longTerm.split(" ")[-1][
            0]):  # note that the position of " " should add 1 because of the earlier program
            # print shortTerm, "," , longTerm
            return False
        else:
            return True
    else:
        return False


# check if one term is an abbreviation of the other
def __isAbrreviation(longTerm, shortTerm):
    try:
        longTerm.encode(encoding="ascii")
        shortTerm.encode(encoding="ascii")
    except:
        return False

    longTerm = longTerm.replace("-", " ").replace("_", " ").strip()
    shortTerm = shortTerm.replace("-", " ").replace("_", " ").strip()

    # the length of the abbreviation must be shorter than the full style at least 2 letters
    if (len(longTerm) - len(shortTerm) < 2) or len(longTerm) < 1 or len(shortTerm) < 1:
        return False

    # the length of the abbreviation should not be so long compared to its postential full name
    if float(len(shortTerm)) / len(longTerm) > 0.68 and not __numberInString(
            shortTerm):  # the parameter is determined by observation
        return False

    if len(shortTerm) > 10:  # the short term should not be so long
        return False

    if ("++" in shortTerm) != ("++" in longTerm):  # if they both include "++" or neither do they
        return False

    if " " in shortTerm or " " in longTerm:  # if the parts of short term is part of the long term, it is not regarded as abbreviations
        if set(shortTerm.split(" ")) < set(longTerm.split(" ")):
            return False

    if longTerm[0] == "." and shortTerm[0] == "." and not longTerm.startswith(".net") and not shortTerm.startswith(
            ".net"):  # As term beginning with dot is always the file extension, it always is not the abbreviation
        return False

    if shortTerm[0] != longTerm[0]:  # two terms should have the same first letter
        return False

    if longTerm.split(" ")[0] + "s" == shortTerm:  # e.g., bolg aritcles --> blogs
        return False

    # if any letter in the short term is not contained in the long term, it is not an abbreviation
    for letter in shortTerm:
        if letter != "-" and letter != "_":
            if letter not in longTerm:
                return False
                break

    # check if number in terms are the same
    if __numberInString(longTerm) or __numberInString(shortTerm):
        if not __isNumberSame(longTerm, shortTerm):
            return False

    # check if the letter order is the same
    if __checkLetterOrder(shortTerm, longTerm):
        # if shortTerm in a consecutive part of long term, it is not an abbreviation
        if shortTerm.replace(" ", "") in longTerm.replace(" ", ""):
            return False
        return True
    else:
        return False


# check if two terms are synonyms
def __isSynonym(term1, term2):
    try:
        term1.encode(encoding="ascii")
        term2.encode(encoding="ascii")
    except:
        return False

    if term1 == term2 or term1.replace(" ", "") == term2.replace(" ", ""):
        return True
    # the numbers inside them should be the same
    if not __isNumberSame(term1, term2):
        return False

    if ("++" in term1) != ("++" in term2):  # if they both include "++" or neither do they    like c++
        # print term1, term2
        return False

    # there are some one-letter change which may results in different meaning such as "encode" and "decode"
    if (term1.replace(" ", "").replace("en", "de") == term2.replace(" ", "")) or (
            term1.replace(" ", "").replace("de", "en") == term2.replace(" ", "")):
        # print term1, "," ,term2
        return False

    # the first letter of two terms should be the same
    if term1[0] != term2[0]:
        return False

    if term1[0] == "." and term2[
        0] == ".":  # As term beginning with dot is always the file extension, it always is not the abbreviation
        return False
    try:
        absoluteDis = levenshtein_distance(term1, term2)  # absolute edit distance
    except Exception as e:
        print(e, term1, term2)
        return False
    # relative edit distance by diving the length of two terms23
    relativeDis = float(absoluteDis) / max((len(term1), len(term2)))
    if absoluteDis < 4 and relativeDis <= 0.3:  # set the absoluteDis for long term while relativeDis for short term
        return True
    else:
        return False


def default_classify_func(term1, term2):
    if __isSynonym(term1, term2):
        return "synonym"
    elif __isAbrreviation(term1, term2):
        return "abbreviation"
    else:
        return "other"


def get_default_synonym_types():
    return ["abbreviation", "other", "synonym"]


class WordDiscrimination(object):
    def __init__(self, classify_word_func, semantic_related_types,group_dict=False,group_word_type="synonym",
                 domain_vocab=None):

        """
        :param classify_word_func: function, parameters: term1,term2, return the type of term2 to term1
        :param semantic_related_types: list, all semantic related types
        """
        self.classify_word_func = classify_word_func
        self.synonym_types = semantic_related_types
        self.group_dict = group_dict
        self.group_word_type = group_word_type
        self.domain_vocab = domain_vocab

    def __group_dict(self,dst):
        if not isinstance(self.domain_vocab,dict):
            raise TypeError("type of domain_vocab  should be  dict")
        G = nx.Graph()
        nodes, edges = [], []
        for k, v in dst.items():
            nodes.append(k)
            for i in v[self.group_word_type]:
                edges.append((k, i))
                nodes.append(i)
        G.add_nodes_from(nodes)
        G.add_edges_from(edges)
        groups = []
        for i in nx.connected_component_subgraphs(G):
            groups.append(list(i))
        # get new dict
        newDi = {}
        otherKeys = list(v.keys()).copy()
        otherKeys.remove(self.group_word_type)
        for group in groups:
            key = max(group, key=lambda x: self.domain_vocab[x] if x in self.domain_vocab else 0)
            group.remove(key)
            newDi[key] = {self.group_word_type: group}
            for i in otherKeys:
                newDi[key][i] = []
                for j in newDi[key][self.group_word_type]:
                    if j in dst:
                        newDi[key][i].extend(dst[j][i])
        return newDi

    def discriminate_words(self, vocab):
        """
        classifiy word
        :param vocab: dict, key:term, value: list, the semantic realted words of this term

        :return: dict, key:term, value:dict(key:synonym type, value: list, words of this synonym type )
        """
        res = {}
        for k, v in vocab.items():
            res[k] = {}
            for i in self.synonym_types:
                res[k][i] = []
            for i in v:
                res[k][self.classify_word_func(k, i)].append(i)
        if self.group_dict:
            logging.info("group thesaurus......")
            return self.__group_dict(res)
        else:
            return res


if __name__ == "__main__":
    pass



import pytest
from os.path import dirname
import sys
sys.path.append(dirname(dirname(__file__)))
from CleanData import CleanDataWiki


def test_wrong_wiki_path():
    cdw = CleanDataWiki("","")
    cdw.tranform()


from collections import defaultdict
def corpusToVocab(sentences):
    vocab = defaultdict(int)
    for line in sentences:
        for word in line.split():
            vocab[word] += 1
    return vocab



import re


def filterTerm(term):
    try:
        term.encode(encoding="ascii")
    except:
        return False

    if len(term) == 1 and term != "r" and term != "c":
        return False
    if term[-1] == "." or term[-1] == "_":
        return False
    # the term cannot include these punctuations
    special_list = ["@", "*", "[", "]", "=", ">", "<", "\\", "/", "|", "&", "$", "~", "^", "%", "..", "--", "__", "`",
                    "´", "’", "‘", "“", "”", " ", "£", "¿", "www.", ".com", "test.", "slow", "'s", ".org", ".axd",
                    ".app",
                    ".mail", ".test", ".mf", ".location",
                    ".system", ".imageio", ".io", ".beginform", ".futures", ".mappath", ".drawing", ".observablearray",
                    ".conf", ".table",
                    ".path", ".actionlink", ".settings", ".log", ".util", ".concurrent"]
    for i in special_list:
        if i in term:
            return False
            return False
    # remove terms whose begining is
    firstLetter_set = set(["#", "_", "+", "-"])
    if term[0] in firstLetter_set:
        return False
    # remove terms whose first character is a digit except some special cases
    if len(re.findall(r"\d\.\d", term)) > 0:  # no version number such as "3.2.1"
        return False

    signal = False  # fail if there are no letters in the term
    for character in term:
        if character.isalpha():
            signal = True
            break
    return signal



import sys
def _check_type(s):
    if sys.version_info[0] == 3 and not isinstance(s, str):
        raise TypeError('expected str or unicode, got %s' % type(s).__name__)
    elif not sys.version_info[0] == 3 and not isinstance(s, unicode):
        raise TypeError('expected unicode, got %s' % type(s).__name__)



def levenshtein_distance(s1, s2):
    """
    compute levenshtein distance
    :param s1: str
    :param s2: str
    :return: levenshtein_distance
    """
    _check_type(s1)
    _check_type(s2)

    if s1 == s2:
        return 0
    rows = len(s1)+1
    cols = len(s2)+1

    if not s1:
        return cols-1
    if not s2:
        return rows-1

    prev = None
    cur = range(cols)
    for r in range(1, rows):
        prev, cur = cur, [r] + [0]*(cols-1)
        for c in range(1, cols):
            deletion = prev[c] + 1
            insertion = cur[c-1] + 1
            edit = prev[c-1] + (0 if s1[r-1] == s2[c-1] else 1)
            cur[c] = min(edit, deletion, insertion)

    return cur[-1]
def StrSimilarity(s1,s2):
    return levenshtein_distance(s1,s2)/max((len(s1),len(s2)))
if __name__ =="__main__":
    print(StrSimilarity("123","1234"))


import requests
import zipfile
import warnings
from sys import stdout
from os import makedirs
from os.path import dirname
from os.path import exists
import logging

logger = logging.getLogger(__name__)

class GoogleDriveDownloader:
    """
    Minimal class to download shared files from Google Drive.
    """

    CHUNK_SIZE = 32768
    DOWNLOAD_URL = "https://docs.google.com/uc?export=download"

    @staticmethod
    def download_file_from_google_drive(file_id, dest_path, overwrite=False, unzip=False):
        """
        Downloads a shared file from google drive into a given folder.
        Optionally unzips it.
        Parameters
        ----------
        file_id: str
            the file identifier.
            You can obtain it from the sharable link.
        dest_path: str
            the destination where to save the downloaded file.
            Must be a path (for example: './downloaded_file.txt')
        overwrite: bool
            optional, if True forces re-download and overwrite.
        unzip: bool
            optional, if True unzips a file.
            If the file is not a zip file, ignores it.
        Returns
        -------
        None
        """

        destination_directory = dirname(dest_path)
        if not exists(destination_directory):
            makedirs(destination_directory)

        if not exists(dest_path) or overwrite:

            session = requests.Session()

            logger.info('Downloading {} into {}... '.format(file_id, dest_path))
            stdout.flush()

            response = session.get(GoogleDriveDownloader.DOWNLOAD_URL, params={'id': file_id}, stream=True)

            token = GoogleDriveDownloader._get_confirm_token(response)
            if token:
                params = {'id': file_id, 'confirm': token}
                response = session.get(GoogleDriveDownloader.DOWNLOAD_URL, params=params, stream=True)

            GoogleDriveDownloader._save_response_content(response, dest_path)
            logger.info('Finish download')

            if unzip:
                try:
                    logger.info('Unzipping...')
                    stdout.flush()
                    with zipfile.ZipFile(dest_path, 'r') as z:
                        z.extractall(destination_directory)
                        logger.info('Finish unzip.')
                except zipfile.BadZipfile:
                    warnings.warn('Ignoring `unzip` since "{}" does not look like a valid zip file'.format(file_id))

    @staticmethod
    def _get_confirm_token(response):
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                return value
        return None

    @staticmethod
    def _save_response_content(response, destination):
        with open(destination, "wb") as f:
            for chunk in response.iter_content(GoogleDriveDownloader.CHUNK_SIZE):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)


class DownloadData():
    def __init__(self):
        self.fileId = {
            "test": "17uxLDQu1tOqHG6XZ7AiKabvAIY-iwJnF",
            "general_vocab": "1SJonZ_-ns8HKAqsHCgswHssXlUNyX9gW",
            "eng_corpus": "1ez1riFK_EDNfnHBwK7zLhXLUcYD7uYTh",
        }

    def download_data(self, dest_path, download_file_name="test", overwrite=False):
        GoogleDriveDownloader.download_file_from_google_drive(file_id=self.fileId[download_file_name], dest_path=dest_path,
                                                              overwrite=overwrite, unzip=True)


if __name__ == "__main__":
    pass



"""
class to get semantic related words for terms
"""
import os
import logging

logger = logging.getLogger(__name__)
from gensim.models import Word2Vec, FastText
from gensim.models.word2vec import LineSentence


class SemanticRelatedWord(object):
    """
    class to get semantic related words for terms.
    This class use word2vec and fasttext.
    """

    def __init__(self, domain_corpus_phrase_path, fasttext_path, skipgram_path, file_overwrite=False,
                 topn_fasttext=8, topn_skipgram=15, similarity_threshold_fasttext=0.8,
                 similarity_threshold_skipgram=0.78, min_count=5, size=200, workers=8, window=5
                 ):
        """
        :param domain_corpus_phrase_path: str, path of domain corpus
        :param fasttext_path: str, the path of fasttext to be saved or the path of existing fasttext model
        :param skipgram_path: tstr, he path of skipgram to be saved or the path of existing skipgram model
        :param file_overwrite: str, if the fasttext or skipgram is already existing, whether re-train
        :param topn_fasttext: int, optional,the number of most similar words chosen by fasttext
        :param topn_skipgram: int, optional, the number of most similar words chosen by fasttext
        :param min_count: int, optional
            The model ignores all words with total frequency lower than this.
        :param size: int, optional
            Dimensionality of the word vectors.
        :param workers: int, optional
            Use these many worker threads to train the model (=faster training with multicore machines).
        :param window: int, optional
            The maximum distance between the current and predicted word within a sentence.
        """
        self.domain_corpus_phrase_path = domain_corpus_phrase_path
        self.fasttext_path = fasttext_path
        self.skipgram_path = skipgram_path
        self.file_overwrite = file_overwrite
        self.topn_fasttext = topn_fasttext
        self.topn_skipgram = topn_skipgram
        self.min_count = min_count
        self.size = size
        self.workers = workers
        self.window = window
        self.similarity_threshold_fasttext = similarity_threshold_fasttext
        self.similarity_threshold_skipgram = similarity_threshold_skipgram

    def getSemanticRelatedWords(self, terms):
        """
        get semantic related words
        :param terms: list, collections of terms
        :return: dict, key is term, value is the SemanticRelatedWords of this term, stored in a list
        """
        # get fasttext and skipgram models
        if os.path.exists(self.fasttext_path) and self.file_overwrite == False:  # exist, just load
            logger.warning(self.fasttext_path + " already exists, program will load it.")
            self.fasttext = FastText.load(self.fasttext_path)
        else:  # train
            logger.info("train fasttext...")
            if not os.path.exists(os.path.dirname(self.fasttext_path)):
                os.makedirs(os.path.dirname(self.fasttext_path))
            self.fasttext = FastText(sentences=LineSentence(self.domain_corpus_phrase_path),
                                     min_count=self.min_count, size=self.size, sg=1, workers=self.workers,
                                     window=self.window)
            logger.info("save fasttext to local...")
            self.fasttext.save(self.fasttext_path)

        if os.path.exists(self.skipgram_path) and self.file_overwrite == False:  # exist, just load
            logger.warning(self.skipgram_path + " already exists, program will load it.")
            self.skipgram = Word2Vec.load(self.skipgram_path)
        else:  # not exist, need to train
            logger.info("train skipgram...")
            if not os.path.exists(os.path.dirname(self.skipgram_path)):
                os.makedirs(os.path.dirname(self.skipgram_path))
            self.skipgram = Word2Vec(
                sentences=LineSentence(self.domain_corpus_phrase_path),
                min_count=self.min_count, size=self.size, sg=1, workers=self.workers, window=self.window)
            self.skipgram.delete_temporary_training_data(True)
            logger.info("save skipgram to local...")
            self.skipgram.save(self.skipgram_path)
        # get semantic related words
        res = {}
        for term in terms:
            try:
                words_fasttext = self.fasttext.wv.most_similar(term, topn=self.topn_fasttext)
            except:
                words_fasttext = []
            try:
                words_skipgram = self.skipgram.wv.most_similar(term, topn=self.topn_skipgram)
            except:
                words_skipgram = []
            words1, words2 = [], []
            for i in words_fasttext:
                if i[1] < self.similarity_threshold_fasttext:
                    break
                words1.append(i[0])
            for i in words_skipgram:
                if i[1] < self.similarity_threshold_skipgram:
                    break
                words2.append(i[0])
            res[term] = list(set(words1 + words2))
        return res



from . import datasets
from . import domain_term
from . import domain_thesaurus
from . import phrase_detection
from . import semantic_related_word
from . import utils
from . import word_discrimination




"""
class to detect phrase
"""
import codecs
import gc
import logging
import os

logger = logging.getLogger(__name__)
from gensim.models.phrases import Phrases, Phraser
from gensim.models.word2vec import LineSentence


class TxtIter(object):
    def __init__(self, sentences, ngrams):
        self.ngrams = ngrams
        self.sentences = sentences

    def __iter__(self):
        if len(self.ngrams) == 0:
            c = 0
            for line in self.sentences:
                c += 1
                if c % 10000 == 0:
                    logger.info("already read {}w lines".format(str(c / 10000)))
                yield line.split()
            self.sentences.close()
        else:
            c = 0
            for line in self.sentences:
                c += 1
                if c % 10000 == 0:
                    logger.info("already read {}w lines".format(str(c / 10000)))
                line = self.ngrams[0][line.split()]
                for gram in self.ngrams[1:]:
                    line = gram[line]
                yield line
            self.sentences.close()


class PhraseDetection(object):
    """
    class to detect phrase
    using Phrases in gensim
    see more details in https://arxiv.org/abs/1310.4546 and https://svn.spraakdata.gu.se/repos/gerlof/pub/www/Docs/npmi-pfd.pdf
    
    """

    def __init__(self, savePhraserPaths, file_overwrite=False, min_count=10, threshold=15.0,
                 max_vocab_size=40000000, delimiter=b'_', scoring='default', wordNumInPhrase=3):
        """
        :param savePhraserPaths: list, the paths of phrases to save
        :param file_overwrite: str, if the phrase is existing, whether overwrite
        :param min_count: float, optional
            Ignore all words and bigrams with total collected count lower than this value.
        :param threshold: float, optional
            Represent a score threshold for forming the phrases (higher means fewer phrases).
            A phrase of words `a` followed by `b` is accepted if the score of the phrase is greater than threshold.
            Heavily depends on concrete scoring-function, see the `scoring` parameter.
        :param max_vocab_size: int, optional
            Maximum size (number of tokens) of the vocabulary. Used to control pruning of less common words,
            to keep memory under control. The default of 40M needs about 3.6GB of RAM. Increase/decrease
            `max_vocab_size` depending on how much available memory you have.
        :param delimiter:str, optional
            Glue character used to join collocation tokens, should be a byte string (e.g. b'_').
        :param scoring:{'default', 'npmi', function}, optional
            Specify how potential phrases are scored. `scoring` can be set with either a string that refers to a
            built-in scoring function, or with a function with the expected parameter names.
            Two built-in scoring functions are available by setting `scoring` to a string:

            #. "default" - :func:`~gensim.models.phrases.original_scorer`.
            #. "npmi" - :func:`~gensim.models.phrases.npmi_scorer`.
        :param wordNumInPhrase: word number in a phrase, Note that `wordNumInPhrase-1` must be equal to length of savePhraserPaths
        """
        self.phrasers = []
        self.savePhraserPaths = savePhraserPaths
        self.file_overwrite = file_overwrite
        self.min_count = min_count
        self.threshold = threshold
        self.max_vocab_size = max_vocab_size
        self.delimiter = delimiter
        self.scoring = scoring
        self.wordNumInPhrase = wordNumInPhrase

    def fit(self, sentencesPath):
        """
        train phrases
        :param sentencesPath:the path of text file, the text file should be the format: one line one sentence
        """
        self.phrasers = []
        # path detect
        for path in self.savePhraserPaths:
            if not os.path.exists(os.path.dirname(path)):
                raise FileNotFoundError(os.path.dirname(path) + " not exist")
        for path in self.savePhraserPaths:
            if not os.path.exists(path):  # need train
                self.phrasers = None
                break
        if self.phrasers is not None and self.file_overwrite == False:
            logging.info("models are already exist, will read it")
            for path in self.savePhraserPaths:
                self.phrasers.append(Phraser(Phrases.load(path)))
            return True
        self.phrasers = []
        c = 2
        for path in self.savePhraserPaths:
            logging.info("getting %d-gram phrase......" % c)
            c += 1
            phrase = Phrases(
                sentences=TxtIter(sentences=codecs.open(sentencesPath, mode="r", encoding="utf-8"),
                                  ngrams=self.phrasers),
                min_count=self.min_count, threshold=self.threshold, max_vocab_size=self.max_vocab_size,
                delimiter=self.delimiter, scoring=self.scoring)
            phrase.save(path)
            phraser = Phraser(phrase)
            self.phrasers.append(phraser)
            del phrase

    def transform(self, sentencesPath, savePath):
        """
        use trained phrases to transform sentences
        :param sentencesPath: the path of text file, the text file should be the format: one line one sentence
        :param savePath: the path of transformed text file, the text file are the format: one line one sentence
        """
        with codecs.open(savePath, mode="w", encoding="utf-8") as fr:
            sentences = TxtIter(sentences=codecs.open(sentencesPath, mode="r", encoding="utf-8"), ngrams=self.phrasers)
            lines = []
            for line in sentences:
                lines.append(" ".join(line) + "\n")
                if len(lines) > 500000:
                    fr.writelines(lines)
                    lines = []
            fr.writelines(lines)
        logger.info("delete all phraser to save memory")
        for i in self.phrasers:
            del i
        del self.phrasers
        self.phrasers = None
        gc.collect()


if __name__ == "__main__":
    ls = LineSentence("E:/a.txt")
    for i in ls:
        print(i)



"""
class to extract domain-specific terms
for examples, 'JavaScript' in CS and 'limit' in math
"""
import logging

logger = logging.getLogger(__name__)
from DST.utils.TermUtil import filterTerm

class DomainTerm(object):
    """
    class to extract domain-specific terms
    """

    def __init__(self, maxTermsCount=300000, thresholdScore=10.0, termFreqRange=(30, float("inf"))):
        """
        :param maxTermsCount: the max number of domain terms
        :param thresholdScore: word larger than thresholdScore will be recognized as domain term
        :param termFreqRange: tuple-like object(minFreq,maxFreq), if word is not in the range, it will be not considered
                as term
        """
        self.maxTermsCount = maxTermsCount
        self.thresholdScore = thresholdScore
        self.termFreqRange = termFreqRange

    def extract_term(self, domainSpecificVocab, generalVocab):
        """
        extract domain term
        :param domainSpecificVocab: all words in domain-corpus, dict object, {"word1":word1Count,"word1":word1Count,...}
        :param generalVocab: all words in general-corpus, dict object, {"word1":word1Count,"word1":word1Count,...}
        :return: terms of list.[term1,term2,term3 ....]
        """
        # get word count in the two vocabulary
        domainSpecificVocabCount, generalVocabCount = 0.0, 0.0
        for _, v in domainSpecificVocab.items():
            domainSpecificVocabCount += v
        for _, v in generalVocab.items():
            generalVocabCount += v
        # extract domain specific terms
        candidateTerms = []
        for word, freq in domainSpecificVocab.items():
            if freq < self.termFreqRange[0] or freq > self.termFreqRange[1] or not filterTerm(word):
                continue
            if word not in generalVocab:
                candidateTerms.append((word, float("inf")))
            else:
                score = (freq / domainSpecificVocabCount) / (generalVocab[word] / generalVocabCount)
                if score > self.thresholdScore:
                    candidateTerms.append((word, score))
        candidateTerms.sort(key=lambda x: x[1], reverse=True)
        terms = candidateTerms[0:self.maxTermsCount]
        logging.info("extract %d terms in total" % len(terms))
        return [term[0] for term in terms]


if __name__ == "__main__":
    print(30, 1 < float("inf"))



"""
to create a domain specific thesaurus by using our approach
the input are domain corpus and gengeral corpus and the output is domain specific thesaurus(dst)

future optimization：
1. some modules' function should be clear and vivid. May remove or add some functions.
2. the program may re-read some data as input, but the output already exist, it wastes some time
3. the information in logging-info should be more clear
4. unit-test (pytest) should be added.
5. use logger more reasonable
6. DST.word_discrimination.word_discrimination is not good
"""
import codecs
import json
import logging
import os

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)
logging.getLogger("gensim").setLevel(logging.WARNING)
logging.getLogger("summarizer").setLevel(logging.WARNING)

from DST.phrase_detection.PhraseDetection import PhraseDetection
from DST.domain_term.DomainTerm import DomainTerm
from DST.semantic_related_word.SemanticRelatedWord import SemanticRelatedWord
from DST.word_discrimination.WordDiscrimination import WordDiscrimination, default_classify_func
from DST.utils.VocabUtil import corpusToVocab


class DomainThesaurus(object):
    """
    class to get domain thesaurus automatically
    """

    def __init__(self,
                 domain_specific_corpus_path,
                 general_vocab_path,
                 outputDir,
                 filePaths=None,
                 phrase_detection_domain="default",
                 domain_specific_term="default",
                 semantic_related_words="default",
                 word_discrimination="default"):
        """
        :param domain_specific_corpus_path: str, the path of domain specific corpus, should be a text file with one line
                 one sentence.
        :param general_vocab_path: str, the path of general vocabulary, should be a json file. Key refer to word and
                value is the count of word.
        :param outputDir: str, the directory of output, the final thesaurus and temporary files  are in the directory

        :param filePaths: dict, optional, specify every path of temporary files. If this parameter is None, filePaths will be generated
                by using outputDir. If the parameter outputDir is None too, the current directory will be the outputDir.
                We suggested that you should specify the ouptputDir or filePaths.

                If you specify filePaths, this dict should have these keys:
                    domain_corpus_phrase: domain corpus with phrase detection

                    general_corpus_phrase: general corpus with phrase detection

                    domain_vocab: vocabulary of domain corpus, it contain all words of this corpus

                    general_vocab: vocabulary of general corpus, it contain all words of this corpus

                    domain_terms: domain terms stored in a text file, one line one term

                    fasttext: the path of fasttext model, if you do not use default class SemanticRelatedWord, it will be
                                unnecessary

                    skipgram: the path of skipgram model, if you do not use default class SemanticRelatedWord, it will be
                                unnecessary

                    semantic_related_words: the path of semantic related words, it is a json file

                    origin_thesaurus: the original thesaurus

                    final_thesaurus: the path of final thesaurus
        :param phrase_detection_domain: class to detect phrase for domain corpus, optional,
                if it is specified, it must be a class with the class functions: fit and transform. The functions must have
                 the same parameters and meaning with the default class

        :param phrase_detection_general: class to detect phrase for general corpus, optional,
                if it is specified, it must be a class with the class functions: fit and transform. The functions must have
                 the same parameters and return with the default class

        :param domain_specific_term: class to extract domain specific terms, optional,
                if it is specified. it must be a class with the class functions: extract_term. The function must have
                the same parameters and return with the default class



        :param semantic_related_words: class to get semantic related words, optional,
                if it is specified. it must be a class with the class functions: getSemanticRelatedWords. The function must have
                the same parameters and return with the default class

        :param word_discrimination: class to  classify semantic related words, optional,
                if it is specified. it must be a class with the class functions: discriminate_words. The function must have
                the same parameters and return with the default class

        :param synonym_group: class to extract group synonyms, optional,
                if it is specified. it must be a class with the class functions: group_synonyms. The function must have
                the same parameters and return with the default class
        """

        self.domain_specific_corpus_path = domain_specific_corpus_path
        self.general_vocab_path = general_vocab_path
        self.outputDir = outputDir
        self.filePaths = filePaths
        self.domain_vocab, self.general_vocab = None, None
        self.domain_terms = None
        self.semantic_related_words = None
        self.origin_thesaurus = None
        self.final_thesaurus = None
        if not os.path.exists(domain_specific_corpus_path):
            logger.error(domain_specific_corpus_path+" does not exists")
            exit(1)
        if not os.path.exists(general_vocab_path):
            logger.error(general_vocab_path+" does not exists")
            exit(1)
        if not os.path.exists(outputDir):
            logger.error(outputDir+" does not exists")
            exit(1)
        self.__getFilePaths()  # get all file paths
        # some models in DST
        if phrase_detection_domain == "default":
            wordNumInPhrase = 3
            savePhraserPaths = []
            phrasesDir = os.path.join(self.outputDir, "Phrases")
            if not os.path.exists(phrasesDir):
                os.makedirs(phrasesDir)
            for i in range(wordNumInPhrase - 1):
                savePhraserPaths.append(
                    os.path.join(self.outputDir, "Phrases/" + str(i + 2) + "_domain_phrase.model"))
            self.PhraseDetectionDomain = PhraseDetection(savePhraserPaths=savePhraserPaths, file_overwrite=False,
                                                         min_count=15, threshold=20, max_vocab_size=40000000,
                                                         delimiter=b'_', scoring='default',
                                                         wordNumInPhrase=wordNumInPhrase)
        else:
            self.PhraseDetectionDomain = phrase_detection_domain

        if domain_specific_term == "default":
            self.DomainTerm = DomainTerm(maxTermsCount=2000, thresholdScore=15.0,
                                         termFreqRange=(100, float("inf")))
        else:
            self.DomainTerm = domain_specific_term
        if semantic_related_words == "default":
            self.SemanticRelatedWords = SemanticRelatedWord(self.filePaths["domain_corpus_phrase"],
                                                            self.filePaths["fasttext"], self.filePaths["skipgram"],
                                                            file_overwrite=False, topn_fasttext=3, topn_skipgram=5,
                                                            similarity_threshold_fasttext=0.8,
                                                            similarity_threshold_skipgram=0.78,
                                                            min_count=20, size=200, workers=8, window=5
                                                            )
        else:
            self.SemanticRelatedWords = semantic_related_words
        if word_discrimination == "default":
            self.word_discrimination = WordDiscrimination(classify_word_func=default_classify_func,
                                                          semantic_related_types=["synonym", "abbreviation", "other"],
                                                          group_dict=True, group_word_type="synonym",
                                                          domain_vocab=self.domain_vocab)
        else:
            self.word_discrimination = word_discrimination

    def clear(self):
        del self.domain_vocab
        del self.general_vocab
        del self.domain_terms
        del self.semantic_related_words
        del self.origin_thesaurus
        del self.final_thesaurus

        self.domain_vocab = None
        self.general_vocab = None
        self.domain_terms = None
        self.semantic_related_words = None
        self.origin_thesaurus = None
        self.final_thesaurus = None

    def __getFilePaths(self):
        # ouputDir is a directory that contain many temp file and final thesaurus in the process of extracting thesaurus
        if self.filePaths is None:  # use default setting
            self.filePaths = {}
            self.filePaths["domain_corpus_phrase"] = os.path.join(self.outputDir, "domain_corpus_phrase.txt")
            self.filePaths["domain_vocab"] = os.path.join(self.outputDir, "domain_vocab.json")
            self.filePaths["domain_terms"] = os.path.join(self.outputDir, "domain_terms.txt")
            self.filePaths["fasttext"] = os.path.join(self.outputDir, "fasttext/fasttext.model")
            self.filePaths["skipgram"] = os.path.join(self.outputDir, "skipgram/skipgram.model")
            self.filePaths["semantic_related_words"] = os.path.join(self.outputDir, "semantic_related_words.json")
            self.filePaths["final_thesaurus"] = os.path.join(self.outputDir, "final_thesaurus.json")

    def __phraseDetection(self):
        if not os.path.exists(self.filePaths["domain_corpus_phrase"]):
            logger.info("detect phrase for domain specific corpus...")
            self.PhraseDetectionDomain.fit(sentencesPath=self.domain_specific_corpus_path)
            self.PhraseDetectionDomain.transform(sentencesPath=self.domain_specific_corpus_path,
                                                 savePath=self.filePaths["domain_corpus_phrase"])
        else:
            logger.warning(self.filePaths["domain_corpus_phrase"] + "already exist, program will not detect phrase.")

    def __corpusVocab(self):
        if not os.path.exists(self.filePaths["domain_vocab"]):
            logger.info("get vocabulary from domain corpus......")
            with codecs.open(self.filePaths["domain_corpus_phrase"], mode="r", encoding="utf-8") as fr:
                self.domain_vocab = corpusToVocab(fr)
                logger.info("save domain vocabulary to local......")
                with codecs.open(self.filePaths["domain_vocab"], mode="w", encoding="utf-8") as fw:
                    fw.write(json.dumps(self.domain_vocab))
        else:
            logger.warning(self.filePaths["domain_vocab"] + " already exists, program will read it")
            with codecs.open(self.filePaths["domain_vocab"], mode="r", encoding="utf-8") as fr:
                self.domain_vocab = json.loads(fr.read())
        logger.info("loading general vocabulary from "+self.general_vocab_path+"......")
        with codecs.open(self.general_vocab_path, mode="r", encoding="utf-8") as fr:
            self.general_vocab = json.loads(fr.read())

    def __domainTerm(self):
        if not os.path.exists(self.filePaths["domain_terms"]):
            self.domain_terms = self.DomainTerm.extract_term(domainSpecificVocab=self.domain_vocab,
                                                             generalVocab=self.general_vocab)
            logger.info("save domain terms to local")
            with codecs.open(self.filePaths["domain_terms"], mode="w", encoding="utf-8") as fw:
                fw.writelines([term + "\n" for term in self.domain_terms])
        else:
            logger.warning(self.filePaths["domain_terms"] + " alread exists, program will read it")
            with codecs.open(self.filePaths["domain_terms"], mode="r", encoding="utf-8") as fr:
                self.domain_terms = [line.strip() for line in fr.readlines()]

    def __semanticRelatedWords(self):
        if not os.path.exists(self.filePaths["semantic_related_words"]):
            self.semantic_related_words = self.SemanticRelatedWords.getSemanticRelatedWords(terms=self.domain_terms)
            # save semantic_related_words
            logger.info("save semantic related words to local...")
            with codecs.open(self.filePaths["semantic_related_words"], mode="w", encoding="utf-8") as fw:
                fw.write(json.dumps(self.semantic_related_words))
        else:
            logger.warning(self.filePaths["semantic_related_words"] + "already exists, program will read it")
            with codecs.open(self.filePaths["semantic_related_words"], mode="r", encoding="utf-8") as fr:
                self.semantic_related_words = json.loads(fr.read())

    def __discriminateWords(self):
        if not os.path.exists(self.filePaths["final_thesaurus"]):
            if isinstance(self.word_discrimination, WordDiscrimination):
                if self.word_discrimination.group_dict:
                    self.word_discrimination.domain_vocab = self.domain_vocab
            self.final_thesaurus = self.word_discrimination.discriminate_words(vocab=self.semantic_related_words)
            # save origin_thesaurus
            logger.info("save thesaurus to local")
            with codecs.open(self.filePaths["final_thesaurus"], mode="w", encoding="utf-8") as fw:
                fw.write(json.dumps(self.origin_thesaurus))
        else:
            logger.warning(self.filePaths["final_thesaurus"] + "already exists, program will read it")
            with codecs.open(self.filePaths["final_thesaurus"], mode="r", encoding="utf-8") as fr:
                self.final_thesaurus = json.loads(fr.read())

    def extract(self):
        """
        get domain thesaurus automatically
        Warning: this may take a long time, the time depends on the size of corpus and the CPU's performance .
        Take Stack Overflow as an example,t may take several days.
        When the program finish a part job( can be watched by logging info), you can stop. If you do not delete the
        temporary files, it can continue this job.
        :return: the final thesaurus
        """
        # phrase detection
        logger.info("begin to detect phrase...")
        self.__phraseDetection()
        logger.info("finish phrase detection")
        # extract domain specific term
        # get vocab
        logger.info("get vocabulary from corpus...")
        self.__corpusVocab()
        logger.info("finish get vocabulary from corpus")
        # get domain terms
        logger.info("get domain term...")
        self.__domainTerm()
        logger.info("finish get domain terms")
        # get semantic related words
        logger.info("get semantic related words...")
        self.__semanticRelatedWords()
        logger.info("finish get semantic related words")
        # classify words
        logger.info("discriminate semantic related words...")
        self.__discriminateWords()
        logger.info("finish discriminate semantic related words")
        logger.info("already get domain-specific thesaurus, it is saved in " + self.filePaths["final_thesaurus"])
        return self.final_thesaurus


if __name__ == "__main__":
    pass



from setuptools import setup, find_packages
setup(
    name='DomainThesaurus',
    version="1.1.7",
    description=('extract domain thesaurus automatically'),
    long_description=open('README.rst').read(),
    author='ZhangDun, ChenXiang, Chunyang Chen',
    author_email='dunnzhang0@gmail.com, xchencs@ntu.edu.cn, chunyang.chen@monash.edu',
    maintainer='ZhangDun',
    maintainer_email='dunnzhang0@gmail.com',
    license='MIT',
    packages=find_packages(),
    url='https://github.com/DunZhang/DomainSpecificThesaurus',
    install_requires=['gensim>=3.6.0', 'networkx>=2.1'],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
)



"""
Class  and functions to clean data.
We provide four methods to seperately clean StackOver flow, Mathematics Stack Exchange, English
Language & Usage Stack Exchange and Wikipedia
"""
import codecs
import re
import logging
import pymongo
import regex
logger = logging.getLogger(__name__)
from lxml import etree
from bs4 import BeautifulSoup
def word_token(sent):
    words = sent.split()
    res = []
    for word in words:
        if word[0] in ["'"]:
            word = word[1:]
        if len(word)<1:
            continue
        if word[-2:] == "s'":
            res.append(word[0:-1])
            res.append("'s")
        elif word[-2:] == "'s":
            res.append(word[0:-2])
            res.append("'s")
        elif word[-1] in ["'"]:
            res.append(word[0:-1])
        else:
            res.append(word)
    return res

class CleanDataWiki(object):
    """
    class to clean StackOver flow data
    """

    def __init__(self, wiki_paths, clean_data_path):
        self.__reSub0 = re.compile("(https?|ftp|file)://[-A-Za-z0-9+&@#/%?=~_|!:,.;]+[-A-Za-z0-9+&@#/%=~_|]")  # URL
        self.__reSub1 = re.compile("[\[\]<>`~$\^&*\"=|%@(){}/\\\\]")  # replace with " "
        self.__bracket = r"\((?:[^()]++|(?R))*+\)"
        self.wiki_paths = wiki_paths
        self.clean_data_path = clean_data_path

    def __clean_data(self, strText):
        sentences = []
        strText = strText.lower()
        strText = regex.subf(self.__bracket, "", strText)
        strText = re.sub(self.__reSub0, " ", strText)
        strText = re.sub(self.__reSub1, " ", strText)

        for sentence in re.split("\.[^a-z0-9]|[,:?!;\n\r]", strText):
            word_sen = word_token(sentence)
            if (len(word_sen) > 2):
                sentences.append(u" ".join(word_sen) + "\n")
        return sentences

    def transform(self, batch_size=100000):
        data = []
        fw = open(self.clean_data_path, "w", encoding="utf-8")
        for path in self.wiki_paths:
            print(path)
            with open(path, "r", encoding="utf-8") as fr:
                for line in fr:
                    if line.startswith("<doc") or line.startswith("</doc>") or " " not in line:
                        continue
                    data.extend(self.__clean_data(line))
                    if len(data) > batch_size:
                        fw.writelines(data)
                        data = []
        fw.writelines(data)
        fw.close()
class CleanDataSO_BERT(object):
    """
    class to clean StackOver flow data
    """

    def __init__(self, so_xml_path, clean_data_path, connStr):
        """
        :param so_xml_path: the path of StackOver flow data, the data is big and
                can be downloaded from https://archive.org/download/stackexchange
        :param clean_data_path: save clean data to text file ( one line one sentence).
        """
        usage_scenario = "embedding"
        self.connStr = connStr
        self.usage_scenario = usage_scenario
        if usage_scenario == "embedding":
            self.__reSub0 = re.compile("(https?|ftp|file)://[-A-Za-z0-9+&@#/%?=~_|!:,.;]+[-A-Za-z0-9+&@#/%=~_|]")  # URL
            self.__reSub1 = re.compile("[\[\]<>`~$\^&*\"=|%@(){},:/\\\\]")  # replace with " "
            self.__resplit = re.compile("\.[^a-z0-9]|[?!;\n\r]")
            self.__rePlus = re.compile("[^+]\+[^+]")
            self.__minNum = 4

        elif usage_scenario == "phrase":
            self.__reSub0 = re.compile("(https?|ftp|file)://[-A-Za-z0-9+&@#/%?=~_|!:,.;]+[-A-Za-z0-9+&@#/%=~_|]")  # URL
            self.__reSub1 = re.compile("[\[\]<>`~$\^&*\"=|%@(){}/\\\\]")  # replace with " "
            self.__resplit = re.compile("\.[^a-z0-9]|[,:?!;\n\r]")
            self.__rePlus = re.compile("[^+]\+[^+]")
            self.__minNum = 2
            self.__bracket = r"\((?:[^()]++|(?R))*+\)"

        self.so_xml_path = so_xml_path
        self.clean_data_path = clean_data_path

    def __clean_data(self, strText):
        sentences = []
        strText = strText.lower()
        if self.usage_scenario == "phrase":
            strText = regex.subf(self.__bracket, "", strText)
        strText = re.sub(self.__reSub0, " ", strText)
        strText = re.sub(self.__reSub1, " ", strText)

        for sub in set(re.findall(self.__rePlus, strText)):
            strText = strText.replace(sub, sub[0] + " " + sub[2])
        for sentence in re.split(self.__resplit, strText):
            word_sen = word_token(sentence)
            if (len(word_sen) > self.__minNum):
                sentences.append(u" ".join(word_sen) + "\n")
        return sentences

    def toMongoDb(self):
        """
        clean data
        """
        logger.info("clean stack overflow data")
        context = etree.iterparse(self.so_xml_path, encoding="utf-8")
        #        fw = codecs.open(self.clean_data_path, mode="w", encoding="utf-8")

        myclient = pymongo.MongoClient(self.connStr)
        mydb = myclient["SODB"]
        mycollection = mydb.create_collection("texts")

        c = 0
        for _, elem in context:  # 迭代每一个
            clean_data = []
            c += 1
            if (c % 100000 == 0):
                logger.info("already clean record:" + str(c / 10000) + "W")
            Id, title, body, typeId, pId = elem.get("Id"), elem.get("Title"), elem.get("Body"), elem.get(
                "PostTypeId"), elem.get("ParentId")
            elem.clear()
            if typeId is None or (int(typeId) != 1 and int(typeId) != 2):
                continue
            if body is not None:
                soup = BeautifulSoup(body, "lxml")
                for pre in soup.find_all("pre"):
                    if (len(pre.find_all("code")) > 0):
                        pre.decompose()
                clean_data.extend(self.__clean_data(soup.get_text()))
            if title is not None:
                clean_data.extend(self.__clean_data(BeautifulSoup(title, "lxml").get_text()))
            if len(clean_data) == 0:
                continue

            if int(typeId) == 1:
                # 直接插入到MongoDB中
                mycollection.insert_one({"Id": int(Id), "data": clean_data})
            elif pId is not None:
                res = mycollection.find_one({"Id": int(pId)})
                if res is not None:
                    # updatae
                    res["data"].extend(clean_data)
                    mycollection.replace_one({"Id": int(pId)}, res)
                else:
                    mycollection.insert_one({"Id": int(pId), "data": clean_data})
        myclient.close()

    def toBertText(self):
        fw = codecs.open(self.clean_data_path, mode="w", encoding="utf-8")
        myclient = pymongo.MongoClient(self.connStr)
        mc = myclient["SODB"]["texts"]
        for i in mc.find():
            fw.writelines(i["data"])
            fw.write("\n")
        fw.close()
        myclient.close()
class CleanDataSO(object):
    """
    class to clean StackOver flow data
    """

    def __init__(self, so_xml_path, clean_data_path):
        """
        :param so_xml_path: the path of StackOver flow data, the data is big and
                can be downloaded from https://archive.org/download/stackexchange
        :param clean_data_path: save clean data to text file ( one line one sentence).
        """
        self.__reSub0 = re.compile("(https?|ftp|file)://[-A-Za-z0-9+&@#/%?=~_|!:,.;]+[-A-Za-z0-9+&@#/%=~_|]")  # URL
        self.__reSub1 = re.compile("[\[\]<>`~$\^&*=|%@(){},:\"/'\\\\]")  # replace with " "
        self.__rePlus = re.compile("[^+]\+[^+]")
        self.__reSplit1 = re.compile("\.[^a-z0-9]|[?!;\n\r]")
        self.so_xml_path = so_xml_path
        self.clean_data_path = clean_data_path

    def __clean_data(self, strText):
        sentences = []
        strText = strText.lower()
        strText = re.sub(self.__reSub0, " ", strText)
        strText = re.sub(self.__reSub1, " ", strText)
        # deal '+', remove single'+', keep two more plus
        for sub in set(re.findall(self.__rePlus, strText)):
            strText = strText.replace(sub, sub[0] + " " + sub[2])
        strText = strText.replace("-", "_")
        for sentence in re.split(self.__reSplit1, strText):
            if (len(sentence.split()) > 6):
                sentence += "\n"
                sentences.append(sentence)
        return sentences

    def transform(self):
        """
        clean data
        """
        logger.info("clean stack overflow data")
        context = etree.iterparse(self.so_xml_path, encoding="utf-8")
        fw = codecs.open(self.clean_data_path, mode="w", encoding="utf-8")

        clean_data = []  # 存储title 和 answers
        c = 0
        for _, elem in context:  # 迭代每一个
            c += 1
            if (c % 100000 == 0):
                logger.info("already clean record:" + str(c / 10000) + "W")
            title, body, typeId = elem.get("Title"), elem.get("Body"), elem.get("PostTypeId")
            elem.clear()
            if typeId is None:
                continue
            if int(typeId) != 1 and int(typeId) != 2:
                continue
            if body is not None:
                soup = BeautifulSoup(body, "lxml")
                for pre in soup.find_all("pre"):
                    if (len(pre.find_all("code")) > 0):
                        pre.decompose()
                clean_data.extend(self.__clean_data(soup.get_text()))
            if title is not None:
                clean_data.extend(self.__clean_data(BeautifulSoup(title, "lxml").get_text()))
            if len(clean_data) > 100000:  # write to local
                fw.writelines(clean_data)
                clean_data = []
        if len(clean_data) > 0:
            fw.writelines(clean_data)
        fw.close()


class CleanDataWiki(object):
    """
    class to clean wiki articles
    """

    def __init__(self, wiki_data_path, clean_data_path):
        """
        :param wiki_data_path: the path of wiki data, the data is big and
                can be downloaded from https://dumps.wikimedia.org/enwiki/latest/
        :param clean_data_path: save clean data to text file ( one line one sentence).
        """
        self.wiki_data_path, self.clean_data_path = wiki_data_path, clean_data_path
        self.__PATTERNS = []
        self.__PATTERNS.append(re.compile("\[\[file[\s\S]*?\]\][\r\n]"))  # 匹配[[file开始的段落
        self.__PATTERNS.append(re.compile("\[\[image[\s\S]*?\]\][\r\n]"))  # 匹配[[image开始的段落
        self.__PATTERNS.append(re.compile("\[\[category[\s\S]*?\]\][\r\n]"))  # 匹配[[category开始的段落
        self.__PATTERNS.append(re.compile(r"{{[\s\S]*?}}"))  # filter {{}} 最小匹配
        self.__PATTERNS.append(
            re.compile("(https?|ftp|file)://[-A-Za-z0-9+&@#/%?=~_|!:,.;]+[-A-Za-z0-9+&@#/%=~_|]"))  # filter URL
        self.__PATTERNS.append(re.compile("\[\[[^]]*?\|"))  # 过滤[[ content|
        self.__PATTERNS.append("[']{2,3}")  # ''  '''
        self.__PATTERNS.append(re.compile("[\[\]<>`~$\^&*=|%@(){},:\"/'\\\\]"))  # 一些替换
        #
        self.__PATTERNS_SECT = []
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} footnotes [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} endnotes [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} references [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} external links [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} 	criticisms [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} 	see also [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} further reading [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} sources [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} bibliography [=]{2,8}"))
        self.__PATTERNS_SECT.append(re.compile("[=]{2,8} publications [=]{2,8}"))

    def __clean_data(self, content):
        words = []
        # Step1 获取HTML文本
        content = content.strip().lower()
        content = BeautifulSoup(content, "lxml").get_text()
        # Step 2正则表达式过滤噪音
        # 先过滤没必要的section
        ins = []
        for pat in self.__PATTERNS_SECT:
            patt = re.search(pat, content)
            if (patt is not None):
                ins.append(patt.span()[0])
        if (len(ins) > 0):
            content = content[0:min(ins)]
        re5 = re.compile("\]")  # 不使用空格替换
        reSplit1 = re.compile("\.[^a-z0-9]|[?!;\n\r]")
        for pat in self.__PATTERNS:
            content = re.sub(pat, " ", content)
        content = re.sub(re5, "", content)
        # 分句分词
        content = content.replace("-", "_")
        for sent in re.split(reSplit1, content):
            sen_word = sent.split()
            if (len(sen_word) > 4):
                if (sen_word[0] != "|" and sen_word[0] != "*"):
                    sen_word.append("\n")
                    words.append(" ".join(sen_word))
        return words

    def tranform(self):
        context = etree.iterparse(self.wiki_data_path, encoding="utf-8")
        fw = codecs.open(self.clean_data_path, mode="w", encoding="utf-8")
        clean_data = []
        c = 0
        for _, elem in context:
            tag, content = elem.tag, elem.text
            elem.clear()
            # print(elem)
            if tag is None or content is None or str(tag)[-4:] != "text" or len(content) < 200:
                continue
            c += 1
            if c % 10000 == 0:
                logger.info("already clean passages:" + str(c / 10000) + "W")
            clean_data.extend(self.__clean_data(content=content))
            if len(clean_data) > 100000:  # write to local
                fw.writelines(clean_data)
                clean_data = []
        if len(clean_data) > 0:
            fw.writelines(clean_data)
        fw.close()


def cleanMathXml(xmlPath="../data/Posts.xml", savePath="../result/cleanMath.txt"):
    """
    :param xmlPath: the path of Mathematics Stack Exchange data, the data is big and
                can be downloaded from https://archive.org/download/stackexchange
    :param savePath: save clean data to text file ( one line one sentence).
    :return:
    """
    counts = [0, 0, 0]
    reSub0 = re.compile("(https?|ftp|file)://[-A-Za-z0-9+&@#/%?=~_|!:,.;]+[-A-Za-z0-9+&@#/%=~_|]")  # URL
    reSub1 = re.compile("[,/'\[\]\\\\|}{:\"><~`@#$%^&*()_+=}]")  # replace with " "
    reSplit1 = re.compile("[.?!;\n\r]")
    sentences = []
    context = etree.iterparse(xmlPath, encoding="utf-8")
    datas = []  # 存储title 和 answers
    c = 0
    for _, elem in context:  # 迭代每一个
        c += 1
        if (c % 50000 == 0):
            logger.info("already pasrse record:" + str(c / 10000) + "W")
        title, body, typeId = elem.get("Title"), elem.get("Body"), elem.get("PostTypeId")
        elem.clear()
        if (typeId is None):
            continue
        if (int(typeId) != 1 and int(typeId) != 2):
            continue
        counts[int(typeId)] += 1
        if (body is not None):
            soup = BeautifulSoup(body, "lxml")
            # 去除长短公式证明
            for p in soup.find_all("p"):
                if p.get_text().startswith("$$"):
                    p.decompose()
            datas.append(soup.get_text())
        if (title is not None):
            datas.append(BeautifulSoup(title, "lxml").get_text())
        # 开始处理获取的数据
    logger.info("start to clean data")
    fw = codecs.open(savePath, "w", encoding="utf-8")
    for strText in datas:
        strText = strText.lower()
        strText = re.sub(reSub0, " ", strText)
        strText = re.sub(reSub1, " ", strText)

        strText = strText.replace("-", "_")
        for sentence in re.split(reSplit1, strText):
            if (len(sentence.split()) > 6):
                sentence += "\n"
                sentences.append(sentence)
        if (len(sentences) > 500000):
            fw.writelines(sentences)
            sentences = []
    if (len(sentences) > 0):
        fw.writelines(sentences)
    fw.close()
    return counts


def cleanEngXml(xmlPath="Posts.xml", savePath="cleanEng.txt"):
    """
    :param xmlPath: the path of English Language & Usage Stack Exchange data, the data is big and
                can be downloaded from https://archive.org/download/stackexchange
    :param savePath: save clean data to text file ( one line one sentence).
    :return:
    """
    counts = [0, 0, 0]
    reSub0 = re.compile("(https?|ftp|file)://[-A-Za-z0-9+&@#/%?=~_|!:,.;]+[-A-Za-z0-9+&@#/%=~_|]")  # URL
    reSub1 = re.compile("[\[\]<>`~$\^&*+=|%@(){},:\"/'\\\\]")  # replace with " "
    reSplit1 = re.compile("[.?!;\n\r]")
    sentences = []
    context = etree.iterparse(xmlPath, encoding="utf-8")
    datas = []  # 存储title 和 answers
    c = 0
    for _, elem in context:  # 迭代每一个
        c += 1
        if (c % 50000 == 0):
            logging.info("already pasrse record:", str(c / 10000) + "W")
        title, body, typeId = elem.get("Title"), elem.get("Body"), elem.get("PostTypeId")
        elem.clear()
        if (typeId is None):
            continue
        if (int(typeId) != 1 and int(typeId) != 2):
            continue
        counts[int(typeId)] += 1
        if (body is not None):
            soup = BeautifulSoup(body, "lxml")
            #            for pre in soup.find_all("pre"):
            #                if (len(pre.find_all("code")) > 0):
            #                    pre.decompose()
            datas.append(soup.get_text())
        if (title is not None):
            datas.append(BeautifulSoup(title, "lxml").get_text())
        # 开始处理获取的数据
    logging.info("clean data")
    fw = codecs.open(savePath, "w", encoding="utf-8")
    for strText in datas:
        strText = strText.lower()
        strText = re.sub(reSub0, " ", strText)
        strText = re.sub(reSub1, " ", strText)

        strText = strText.replace("-", "_")
        for sentence in re.split(reSplit1, strText):
            if (len(sentence.split()) > 6):
                sentence += "\n"
                sentences.append(sentence)
        if (len(sentences) > 500000):
            fw.writelines(sentences)
            sentences = []
    if (len(sentences) > 0):
        fw.writelines(sentences)
    fw.close()
    return counts


if __name__ == "__main__":
    context = etree.iterparse("E://a.xml", encoding="utf-8")
    datas = []  # 存储title 和 answers
    # context.
    for _, elem in context:  # 迭代每一个
        print(elem.tag + ":" + elem.text)
        elem.clear()
